# =============================================================================
# Phylactery Union Find
# =============================================================================
#
# A static Union Find data structure useful to find disjoint sets and
# connected components.
#
import math
from collections import defaultdict


class UnionFind(object):
    """
    The UnionFind class.

    Args:
        capacity (number): total number of items to store.

    """

    __slots__ = ('capacity', 'cardinalities', 'count', 'parents', 'ranks')

    def __init__(self, capacity):

        # Properties
        self.capacity = capacity
        self.count = capacity
        self.parents = list(range(capacity))
        self.cardinalities = [1] * capacity
        self.ranks = [0] * capacity

    def __len__(self):
        return self.count

    def find(self, x):
        y = x
        parents = self.parents

        while True:
            c = parents[y]

            if y == c:
                break

            y = c

        # Path compression
        while True:
            p = parents[x]

            if p == y:
                break

            x = p

        return y

    def union(self, x, y):
        x_root = self.find(x)
        y_root = self.find(y)

        parents = self.parents
        cardinalities = self.cardinalities
        ranks = self.ranks

        # x & y are already in the same set
        if x_root == y_root:
            return

        self.count -= 1

        # x & y are not in the same set, we merge them
        x_rank = ranks[x]
        y_rank = ranks[y]

        if x_rank < y_rank:
            cardinalities[y_root] += cardinalities[x_root]
            parents[x_root] = y_root
        elif x_rank > y_rank:
            cardinalities[x_root] += cardinalities[y_root]
            parents[y_root] = x_root
        else:
            cardinalities[x_root] += cardinalities[y_root]
            parents[y_root] = x_root
            ranks[x_root] += 1

    def connected(self, x, y):
        return self.find(x) == self.find(y)

    def cardinality(self, x):
        parent = self.find(x)
        return self.cardinalities[parent]

    def components(self, min_size=1, max_size=float('inf')):

        component_index = defaultdict(list)

        for i in range(self.capacity):
            root = self.find(i)

            cardinality = self.cardinalities[root]

            if cardinality < min_size or cardinality > max_size:
                continue

            component_index[root].append(i)

        yield from component_index.values()

    def __iter__(self):
        return self.components()

    def __getitem__(self, x):
        return self.find(x)

    def __repr__(self):
        return '<%s capacity=%i count=%i>' % (
            self.__class__.__name__,
            self.capacity,
            self.count
        )



# =============================================================================
# Phylactery TrieDict
# =============================================================================
#
# A simple Python implementation of a (key, value) Trie.
#


class TrieDictNode(object):
    __slots__ = ('children', 'value')

    def __init__(self):
        self.children = None
        self.value = None


class TrieDict(object):
    __slots__ = ('__root', '__shape', '__size')

    def __init__(self, shape=str):
        self.__root = TrieDictNode()
        self.__shape = shape
        self.__size = 0

    def __len__(self):
        return self.__size

    def set(self, key, value):
        node = self.__root

        for char in key:

            if node.children is None:
                child = TrieDictNode()

                node.children = {
                    char: child
                }

                node = child
                continue

            child = node.children.get(char)

            if child is not None:
                node = child
            else:
                child = TrieDictNode()

                node.children[char] = child
                node = child

        if node.value is None:
            self.__size += 1

        node.value = value

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key, default=None):
        node = self.__root

        for char in key:
            if node.children is None:
                return default

            child = node.children.get(char)

            if child is None:
                return default

            node = child

        if node.value is not None:
            return node.value

        return default

    def __getitem__(self, key):
        node = self.__root

        for char in key:
            if node.children is None:
                raise KeyError(key)

            child = node.children.get(char)

            if child is None:
                raise KeyError(key)

            node = child

        if node.value is not None:
            return node.value

        raise KeyError(key)

    def longest(self, key):
        node = self.__root

        last_value = None

        for char in key:

            if node.value is not None:
                last_value = node.value

            if node.children is None:
                break

            child = node.children.get(char)

            if child is None:
                break

            node = child

        if node.value is not None:
            return node.value

        return last_value

    def items(self):
        stack = [(self.__root, self.__shape())]

        while len(stack) > 0:
            node, key = stack.pop()

            if node.value:
                yield (key, node.value)

            if node.children is None:
                continue

            for char, child in node.children.items():
                stack.append((child, key + char))

    def keys(self):
        stack = [(self.__root, self.__shape())]

        while len(stack) > 0:
            node, key = stack.pop()

            if node.value:
                yield key

            if node.children is None:
                continue

            for char, child in node.children.items():
                stack.append((child, key + char))

    def values(self):
        stack = [self.__root]

        while len(stack) > 0:
            node = stack.pop()

            if node.value:
                yield node.value

            if node.children is None:
                continue

            stack.extend(node.children.values())

    def __iter__(self):
        return self.items()



# =============================================================================
# Phylactery Vantage Point Tree
# =============================================================================
#
# Python implementation of a Vantage Point Tree (VPTree) that can index items
# in metric space and perform efficient k-nn queries.
#
import math
import random
from statistics import median, pvariance

# TODO: fails if only 1 item given


class VPTreeNode(object):
    """
    Class representing a Vantage Point Tree's node.

    Args:
        value (any): The node's value.

    Attributes:
        left (VPTreeNode): Left child.
        right (VPTreeNode): Right child.
        mu (int or float): Median threshold.
        value (any): The node's value.

    """

    __slots__ = ('left', 'mu', 'right', 'value')

    def __init__(self, value):
        self.value = value
        self.mu = 0.0
        self.left = None
        self.right = None

    def __repr__(self):
        return '<%s value=%s mu=%i>' % (
            self.__class__.__name__,
            self.value,
            self.mu
        )


class VPTree(object):
    """
    Class representing a Vantage Point Tree.

    Args:
        items (iterable): Items to index in the tree.
        distance (callable): Distance function to use.
        selection (str, optional): Vantage point selection mode to choose
            between "arbitrary", "random" and "spread". Defaults to "arbitrary".

    """

    __slots__ = ('count', 'distance', 'root', 'selection')

    def __init__(self, items, distance, selection='arbitrary'):

        if type(items) is not list:
            items = list(items)

        # Properties
        self.distance = distance
        self.selection = selection
        self.count = len(items)

        # Initializing root
        index, indices = self.__select(list(range(self.count)), items)
        self.root = VPTreeNode(items[index])

        # Building the tree
        stack = [(self.root, indices)]

        while len(stack) != 0:
            vp, values = stack.pop()

            # We compute distances from vp to other points
            distances = [0] * len(values)
            for i, v in enumerate(values):
                distances[i] = distance(vp.value, items[v])

            # And we split at median
            mu = median(distances)
            vp.mu = mu

            left = []
            right = []

            for i, v in enumerate(values):
                if distances[i] < mu:
                    left.append(v)
                else:
                    right.append(v)

            if len(right):
                index, right = self.__select(right, items)
                node = VPTreeNode(items[index])
                vp.right = node

                if len(right):
                    stack.append((node, right))

            if len(left):
                index, left = self.__select(left, items)
                node = VPTreeNode(items[index])
                vp.left = node

                if len(left):
                    stack.append((node, left))

    def __len__(self):
        return self.count

    def __select(self, indices, values):
        selection = self.selection

        if selection == 'arbitrary':
            index = indices.pop()
            return index, indices

        if selection == 'random':
            random.shuffle(indices)
            index = indices.pop()
            return index, indices

        if selection == 'spread':
            distance = self.distance

            s = min(math.ceil(len(indices) * 0.1), 100)

            P = random.sample(indices, s)
            best_spread = float('-inf')
            best_p = None

            for p in P:
                D = random.sample(indices, s)
                M = [0] * s

                for i in range(s):
                    M[i] = distance(values[p], values[D[i]])

                spread = pvariance(M)

                if spread > best_spread:
                    best_spread = spread
                    best_p = p

            indices.pop(indices.index(best_p))

            return best_p, indices

    def dfs(self):
        stack = [(0, self.root)]

        while len(stack) != 0:
            level, node = stack.pop()

            yield level, node

            if node.right:
                stack.append((level + 1, node.right))
            if node.left:
                stack.append((level + 1, node.left))

    def neighbors_in_radius(self, query, radius):
        stack = [self.root]
        distance = self.distance

        while len(stack) != 0:
            node = stack.pop()

            d = distance(node.value, query)

            if d <= radius:
                yield (node.value, d)

            mu = node.mu

            if node.left and d < mu + radius:
                stack.append(node.left)

            if node.right and d >= mu - radius:
                stack.append(node.right)



# =============================================================================
# Phylactery BitSet
# =============================================================================
#
# A simple Python BitSet implementation. It seems it remains slower than a
# cpython set which is a tad disappointing to say the least.
#
import math

MODULO = 0x0000001f


class BitSet(object):
    """
    The BitSet class.

    Args:
        capacity (number): Capacity of the bitset.

    """

    __slots__ = ('capacity', 'integers')

    def __init__(self, capacity):
        true_capacity = math.ceil(capacity / 32)

        # Properties
        self.capacity = capacity
        self.integers = [0] * true_capacity

    def __len__(self):
        return self.capacity

    def get(self, index):

        if index >= self.capacity:
            raise IndexError

        byte = index >> 5
        pos = index & MODULO

        return (self.integers[byte] >> pos) & 1

    def has(self, index):
        return self.get(index) == 1

    def test(self, index):
        return self.has(index)

    def set(self, index, value=1):
        if index >= self.capacity:
            raise IndexError

        byte = index >> 5
        pos = index & MODULO

        if not value:
            self.integers[byte] &= ~(1 << pos)
        else:
            self.integers[byte] |= (1 << pos)

    def update(self, iterable):
        for index in iterable:
            self.set(index)

    def reset(self, index):
        if index >= self.capacity:
            raise IndexError

        byte = index >> 5
        pos = index & MODULO

        self.integers[byte] &= ~(1 << pos)

    def add(self, index):
        return self.set(index)

    def delete(self, index):
        return self.reset(index)

    def __getitem__(self, index):
        return self.get(index)

    def __contains__(self, index):
        return self.has(index)

    def __setitem__(self, index, value):
        return self.set(index, value)

    def __delitem__(self, index):
        return self.reset(index)

    def __repr__(self):
        return '<%s capacity=%i>' % (self.__class__.__name__, self.capacity)
